import ast
import os

class PatternFinder(ast.NodeVisitor):

    def __init__(self):
        """
        Initialize two lists to save patterns.
        """
        self.patternsFunc = set([])
        self.patternsClass = set([])


    def visit_Call(self, nodeCall):
        """
        Be invoked when visiting a node of function call.

        @param node: currently visiting node
        """
        super(PatternFinder, self).generic_visit(nodeCall)
        # Capture assignment like 'f = getattr(...)'.
        if hasattr(nodeCall.func, "func"):
            # In this case, the statement should be
            # 'f = getattr(...)()'.
            nodeCall = nodeCall.func
        # Make sure the function's name is 'getattr'.
        if not hasattr(nodeCall.func, "id"):
            return
        if nodeCall.func.id != "getattr":
            return

        # Capture 'f = getattr(foo, "bar_%s" % baz )' or
        # 'f = getattr(foo, "bar_" + baz )'.
        nodeArgument = nodeCall.args[1]
        if not isinstance(nodeArgument, ast.BinOp):
            return
        operation = nodeArgument.op
        if type(operation) not in [ast.Mod, ast.Add]:
            return
        nodePattern = nodeArgument.left
        if not isinstance(nodePattern, ast.Str):
            return
        pattern = nodePattern.s
        if not ((type(operation) == ast.Add and pattern.endswith("_")) or
                (pattern.count("%s") == 1 and pattern.endswith("_%s"))):
            return
        pattern = pattern.replace("%s", "")
        if pattern[:1].isalpha() and not pattern[:1].islower():
            self.patternsClass.add(pattern)
        else:
            self.patternsFunc.add(pattern)



def findPatternsInFile(codes, patternFinder):
    """
    Find patterns of exceptions in a file.

    @param codes: code of the file to check
    @param patternFinder: a visitor for pattern checking and save results
    """
    tree = ast.parse(codes)
    patternFinder.visit(tree)



def findAllExceptions(pathToCheck):
    """
    Find patterns of exceptions in a file or folder.

    @param patternFinder: a visitor for pattern checking and save results
    @return: patterns of special functions and classes
    """
    finder = PatternFinder()
    if os.path.isfile(pathToCheck):
        with open(pathToCheck) as f:
            findPatternsInFile(f.read(), finder)
    else:
        for path, dirs, files in os.walk(pathToCheck):
            for file in files:
                _, extname = os.path.splitext(file)
                if extname == ".py":
                    pathFile = os.path.join(path, file)
                    with open(pathFile) as f:
                        findPatternsInFile(f.read(), finder)
    return finder.patternsFunc, finder.patternsClass



import re

from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker

from twistedchecker.core.util import isTestModule, moduleNeedsTests


class HeaderChecker(BaseChecker):
    """
    A checker for checking headers.
    Moudles must have a copyright message, a docstring and
    a reference to a test module that contains the bulk of its tests.
    Copyright should like this:
    # Copyright (c) Twisted Matrix Laboratories.
    # See LICENSE for details.
    Reference of test module should like this:
    # -*- test-case-name: <test module> -*-
    """
    msgs = {
        'W9001': ('Missing copyright header',
                  'Used when a module of Twisted has no copyright header.',
                  'missing-copyright-header'),
        'W9002': ('Missing a reference to test module in header',
                  'Used when a module does not contain a reference'
                  ' to test module.', 'missing-test-header'),
    }
    __implements__ = IAstroidChecker
    name = 'header'
    options = ()
    commentsCopyright = (br"# Copyright \(c\) Twisted Matrix Laboratories\.",
                         br"# See LICENSE for details\.")
    patternTestReference = (br"# -\*- test-case-name:"
                            br" (([a-z_][a-z0-9_]*)\.)*[a-z_][a-z0-9_]* -\*-")


    def visit_module(self, node):
        """
        A interface will be called when visiting a module.

        @param node: node of current module
        """
        if not node.file_stream:
            # failed to open the module
            return
        text = node.file_stream.read()
        self._checkCopyright(text, node)
        if not isTestModule(node.name) and moduleNeedsTests:
            self._checkTestReference(text, node)


    def _checkCopyright(self, text, node):
        """
        Check whether the module has copyright header.

        @param text: codes of the module
        @param node: node of the module
        """
        if not re.search(br"%s\s*\n\s*%s" % self.commentsCopyright, text):
            self.add_message('W9001', node=node)


    def _checkTestReference(self, text, node):
        """
        Check whether a reference to its test module is contained.

        @param text: codes of the module
        @param node: node of the module
        """
        if '.test.' in node.name or '.test_' in node.name:
            # Test packages or test modules don't need references to tests.
            return

        if not re.search(self.patternTestReference, text):
            self.add_message('W9002', node=node)



# enable: C0103

class FooClass:
    """
    Just a class.
    """
    def __init__(self):
        """
        Some bad variable names.
        """
        FooBar = None
        Foo = None
        FOO = None
        foo_bar = None



# enable: W9401,W9402

#A comment does not begin with a whitespace.

someVariable = 1 + 2  #  A comment begins with two whitespace.

# a comment begins with a lowercase letter.

# Good comment examples.

# A sentence that spans multiple lines
# doesn't need to have capitalization on second line.

# Here's some code samples:
#  x = x + 1

# Make sure no error occur when checking an empty comment
#

"""
#twisted checker thinks this is a comment and so
1. Comments should begin with one whitespace
2. The first letter of comment should be capitalized

Line with an epydoc C{#} markup.

This is particularly triggered with url fragments
@see U{https://example.com/test#fragment}
"""

#This comment should be reported.

someVariable = 1 + 2 #  So does this one.

class SomeClass(object):
    """
    #twisted checker thinks this is a comment.

    Line with an epydoc C{#} markup.

    Also triggered with url fragments U{https://example.com/test#fragment}
    """
    #But this comment should be reported.

# '\r\n\t' a comment can start with a new lines characters.

var = 1 + 2  # \r\n same for inline comments.

# `literal` is fine at the start.




from twisted.trial import unittest

from twistedchecker.core.exceptionfinder import PatternFinder
from twistedchecker.core.exceptionfinder import findPatternsInFile
from twistedchecker.core.exceptionfinder import findAllExceptions
from twisted.python.filepath import FilePath



class ExceptionFinderTestCase(unittest.TestCase):
    """
    Test for twistedchecker.core.exceptionfinder.
    """

    def test_findPatternsInFile(self):
        """
        Test for twistedchecker.core.exceptionfinder.findPatternsInFile.
        """
        codes = """
obj, something = None, None

f = getattr(obj, "foo_" + something)
g = getattr(obj, "Bar_%s" % something)
getattr(obj, "baz_%s" % something)()
        """
        finder = PatternFinder()
        findPatternsInFile(codes, finder)
        self.assertEqual(finder.patternsFunc, {"foo_", "baz_"})
        self.assertEqual(finder.patternsClass, {"Bar_"})


    def test_findAllExceptions(self):
        """
        Test for twistedchecker.core.exceptionfinder.test_findAllExceptions.
        """
        pathTestFiles = createTestFiles(self.mktemp())
        patternsFunc, patternsClass = findAllExceptions(pathTestFiles)

        self.assertEqual(patternsFunc, {"foo_", "baz_"})
        self.assertEqual(patternsClass, {"Bar_"})



def createTestFiles(tempPath):
    """
    Create test files in file system.

    @tempPath: path of temp directory
    @return: path of test files
    """
    tempDir = FilePath(tempPath)
    tempDir.createDirectory()
    moduleInit = tempDir.child('__init__.py')
    moduleInit.setContent(b"")
    # A module declares exception for function names.
    moduleA = tempDir.child('a.py')
    moduleA.setContent(b"""
obj, something = None, None

func = getattr(obj, "foo_" + something)
getattr(obj, "baz_%s" % something)()
    """)
    # A module declares exception for class names.
    moduleB = tempDir.child('b.py')
    moduleB.setContent(b"""
obj, something = None, None

className = getattr(obj, "Bar_%s" % something)
    """)
    # A module contains invalid names.
    moduleTest = tempDir.child('test.py')
    moduleTest.setContent(b"""
# Not invalid names.
def foo_SOMETHING():
    pass

class Bar_SOMETHING():
    def baz_SOMETHING():
        pass

# Invalid names.
def a_SOMETHING():
    pass

class B_SOMETHING():
    def c_SOMETHING():
        pass
    """)
    return tempDir.path



# -*- test-case-name: twistedchecker.test.test_functionaltests -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Checks for docstrings.
"""

import re
import astroid

from astroid import node_classes, scoped_nodes

from pylint.interfaces import IAstroidChecker, INFERENCE, INFERENCE_FAILURE
from pylint.checkers.base import (DocStringChecker as PylintDocStringChecker,
                                  NO_REQUIRED_DOC_RGX)
from pylint.checkers.utils import has_known_bases


def _isInner(node):
    """
    Determine whether the given node is, at any point in its syntactic
    parentage, defined within a function.

    @param node: The node to inspect.
    @type node: L{logilab.astng.bases.NodeNG}

    @return: a boolean indicating if the given node is defined as an inner
        class or inner function.
    """
    while node:
        node = node.parent
        if isinstance(node, scoped_nodes.FunctionDef):
            return True
    return False



def _getDecoratorsName(node):
    """
    Return a list with names of decorators attached to this node.

    @param node: current node of pylint
    """
    # For setter properties pylint fails so we use a custom code.
    decorators = []
    if not node.decorators:
        return decorators

    for decorator in node.decorators.nodes:
        decorators.append(decorator.as_string())
    return decorators


def _isSetter(node_type, node):
    """
    Determine whether the given node is a setter property.

    @param node_type: The type of the node to inspect.
    @param node: The L{logilab.astng.bases.NodeNG} to inspect.

    @return: a boolean indicating if the given node is a setter.
    """
    if node_type not in ['function', 'method']:
        return False

    for name in _getDecoratorsName(node):
        if '.setter' in name:
            return True
    return False


_counter = iter(range(100))


class DocstringChecker(PylintDocStringChecker):
    """
    A checker for checking docstrings.
    """
    msgs = {
     'W9201': ('The opening/closing of docstring should be on a line '
               'by themselves',
               'Check the opening/closing of a docstring.', 'docstring' + str(next(_counter))),
     'W9202': ('Missing epytext markup @param for argument "%s"',
               'Check the epytext markup @param.', 'docstring' + str(next(_counter))),
     'W9203': ('Missing epytext markup @type for argument "%s"',
               'Check the epytext markup @type.', 'docstring' + str(next(_counter))),
     'W9204': ('Missing epytext markup @return for return value',
               'Check the epytext markup @return.', 'docstring' + str(next(_counter))),
     'W9205': ('Missing epytext markup @rtype for return value',
               'Check the epytext markup @rtype.', 'docstring' + str(next(_counter))),
     'W9206': ('Docstring should have consistent indentations',
               'Check indentations of docstring.', 'docstring' + str(next(_counter))),
     'W9207': ('Missing a blank line before epytext markups',
               'Check the blank line before epytext markups.', 'docstring' + str(next(_counter))),
     'W9208': ('Missing docstring',
               'Used when a module, function, class or method '
               'has no docstring.', 'docstring' + str(next(_counter))),
     'W9209': ('Empty docstring',
               'Used when a module, function, class or method '
               'has an empty docstring.', 'docstring' + str(next(_counter))),
    }
    __implements__ = IAstroidChecker
    name = 'docstring'
    options = ()


    def open(self):
        """
        Set a value to stats and config.
        """
        self.stats = None
        self.config.no_docstring_rgx = NO_REQUIRED_DOC_RGX


    def _getLineIndent(self, line):
        """
        Get indentation of a line.

        @param line: a line of code
        @return: number of spaces
        """
        return len(line) - len(line.lstrip(" "))


    def _getDocstringLineno(self, node_type, node):
        """
        Get line number of the docstring.

        @param node_type: type of node_type
        @param node: node of currently checking
        @return: line number
        """
        docstringStriped = node.as_string().strip()
        linenoDocstring = (node.lineno + docstringStriped
                           .count("\n", 0, docstringStriped.index('"""')))
        if node_type == "module":
            # Module starts from line 0.
            linenoDocstring += 1
        return linenoDocstring

    def visit_module(self, node):
        self._check_docstring('module', node)

    def visit_classdef(self, node):
        if self.config.no_docstring_rgx.match(node.name) is None:
            self._check_docstring('class', node)

    def visit_functiondef(self, node):
        if self.config.no_docstring_rgx.match(node.name) is None:
            ftype = node.is_method() and 'method' or 'function'

            if isinstance(node.parent.frame(), astroid.ClassDef):
                overridden = False
                confidence = (INFERENCE if has_known_bases(node.parent.frame())
                              else INFERENCE_FAILURE)
                # check if node is from a method overridden by its ancestor
                for ancestor in node.parent.frame().ancestors():
                    if node.name in ancestor and \
                       isinstance(ancestor[node.name], astroid.FunctionDef):
                        overridden = True
                        break
                self._check_docstring(ftype, node,
                                      report_missing=not overridden,
                                      confidence=confidence)
            else:
                self._check_docstring(ftype, node)


    visit_asyncfunctiondef = visit_functiondef


    def _check_docstring(self, node_type, node, report_missing=True,
                         confidence=None):
        """
        Check whether the opening and the closing of docstring
        on a line by themselves.
        Then check for epytext markups for function or method.

        @param node_type: type of node
        @param node: current node of pylint
        """
        docstring = node.doc
        if docstring is None:
            # The node does not have a docstring.
            if _isInner(node):
                # Do not check things inside a function or method.
                return

            if _isSetter(node_type, node):
                # Setters don't need a docstring as they are documented in
                # the getter.
                return

            self.add_message('W9208', node=node)
            return
        elif not docstring.strip():
            # Empty docstring.
            self.add_message('W9209', node=node)
            return
        # Get line number of docstring.
        linenoDocstring = self._getDocstringLineno(node_type, node)
        self._checkDocstringFormat(node_type, node, linenoDocstring)
        self._checkEpytext(node_type, node, linenoDocstring)
        self._checkBlankLineBeforeEpytext(node_type, node, linenoDocstring)


    def _checkIndentationIssue(self, node, node_type, linenoDocstring):
        """
        Check whether a docstring have consistent indentations.

        @param node: the node currently checks by pylint
        @param node_type: type of given node
        @param linenoDocstring: line number the docstring begins
        """
        indentDocstring = node.col_offset and node.col_offset or 0
        indentDocstring += len(re.findall(r'\n( *)"""',
                                          node.as_string())[0])
        linesDocstring = node.doc.lstrip("\n").split("\n")
        for nline, lineDocstring in enumerate(linesDocstring):
            if (nline < len(linesDocstring) - 1
                and not lineDocstring.strip()):
                # It's a blank line.
                continue
            if indentDocstring != self._getLineIndent(lineDocstring):
                if node_type == "module":
                    lineno = linenoDocstring
                else:
                    lineno = linenoDocstring + nline + 1
                self.add_message('W9206', line=lineno, node=node)


    def _checkDocstringFormat(self, node_type, node, linenoDocstring):
        """
        Check opening/closing of docstring.

        @param node_type: type of node
        @param node: current node of pylint
        @param linenoDocstring: linenumber of docstring
        """
        # Check the opening/closing of docstring.
        docstringStrippedSpaces = node.doc.strip(" ")
        if (not docstringStrippedSpaces.startswith("\n")
            or not docstringStrippedSpaces.endswith("\n")):
            # If the docstring is in one line, then do not check indentations.
            self.add_message('W9201', line=linenoDocstring, node=node)
        else:
            # If the docstring's opening and closing quotes are on separate
            # lines, then we check its indentation.
            # Generating warnings about indentation when the quotes aren't
            # done right only clutters the output.
            self._checkIndentationIssue(node, node_type, linenoDocstring)


    def _hasReturnValue(self, node):
        """
        Determine whether the given method or function has a return statement.

        @param node: the node currently checks
        """
        returnFound = False
        for subnode in node.body:
            if type(subnode) == node_classes.Return and subnode.value:
                returnFound = True
                break
        return returnFound


    def _checkEpytext(self, node_type, node, linenoDocstring):
        """
        Check epytext of docstring.

        @param node_type: type of node
        @param node: current node of pylint
        @param linenoDocstring: linenumber of docstring
        """
        if node_type not in ['function', 'method']:
            return
        # Check for arguments.
        # If current node is method,
        # then first argument could not have a epytext markup.
        # The first argument usually named 'self'.
        argnames = (node.argnames()[1:] if node_type == 'method'
                    else node.argnames())

        if _isSetter(node_type, node):
            # For setter methods we remove the `value` argument as it
            # does not need to be documented.
            try:
                argnames.remove('value')
            except ValueError:
                # No `value` in arguments.
                pass

        for argname in argnames:
            if node.name.startswith('opt_'):
                # The docstring for option methods is presented as user-facing
                # documentation.  Avoid requiring epytext in them.
                return
            if not re.search(r"@param\s+%s\s*:" % argname, node.doc):
                self.add_message('W9202', line=linenoDocstring,
                                 node=node, args=argname)
            if not re.search(r"@type\s+%s\s*:" % argname, node.doc):
                self.add_message('W9203', line=linenoDocstring,
                                 node=node, args=argname)

        self._checkReturnValueEpytext(node, linenoDocstring)


    def _checkReturnValueEpytext(self, node, linenoDocstring):
        """
        Check if return value is documented.

        @param node: current node of pylint
        @param linenoDocstring: linenumber of docstring
        """
        # Getter properties don't need to document their return value,
        # but then need to have a return value.
        if 'property' in _getDecoratorsName(node):
            if self._hasReturnValue(node):
                # Getter properties don't need a docstring.
                return

        # Check for return value.
        if self._hasReturnValue(node):
            if node.name.startswith('test_'):
                # Ignore return documentation for test methods.
                return
            if not re.search(r"@return[s]{0,1}\s*:", node.doc):
                self.add_message('W9204', line=linenoDocstring, node=node)
            if not re.search(r"@rtype\s*:", node.doc):
                self.add_message('W9205', line=linenoDocstring, node=node)


    def _checkBlankLineBeforeEpytext(self, node_type, node, linenoDocstring):
        """
        Check whether there is a blank line before epytext.

        @param node_type: type of node
        @param node: current node of pylint
        @param linenoDocstring: linenumber of docstring
        """
        # Check whether there is a blank line before epytext markups.
        patternEpytext = (r"\n *@(param|type|return|returns|rtype|ivar|cvar"
                          r"|raises|raise)"
                          r"\s*[a-zA-Z0-9_]*\s*\:")
        matchedEpytext = re.search(patternEpytext, node.doc)
        if matchedEpytext:
            # This docstring have epytext markups,
            # then check the blank line before them.
            posEpytext = matchedEpytext.start() + 1
            if not re.search(r"\n\s*\n\s*$", node.doc[:posEpytext]):
                self.add_message('W9207', line=linenoDocstring, node=node)



from twisted.trial import unittest

from twistedchecker.checkers.docstring import DocstringChecker



class DocstringTestCase(unittest.TestCase):
    """
    Test for twistedchecker.checkers.docstring
    """

    def test_getLineIndent(self):
        """
        Test of twistedchecker.checkers.docstring._getLineIndent.
        """
        checker = DocstringChecker()
        indentNoSpace = checker._getLineIndent("foo")
        indentTwoSpaces = checker._getLineIndent("  foo")
        indentFourSpaces = checker._getLineIndent("    foo")
        self.assertEqual(indentNoSpace, 0)
        self.assertEqual(indentTwoSpaces, 2)
        self.assertEqual(indentFourSpaces, 4)



"""
Extension of pylint format checkers.
"""
from pylint.checkers.format import FormatChecker


def check_lines(self, lines, i):
    """
    check lines have less than a maximum number of characters.

    It ignored lines with long URLs.
    """
    maxChars = self.config.max_line_length
    for line in lines.splitlines():
        if len(line) > maxChars:
            if 'http://' in line or 'https://' in line:
                continue
            self.add_message('C0301', line=i, args=(len(line), maxChars))
        i += 1



def patch():
    """
    pylint thinks that its default checkers are so special that anybody
    wants them so there is no clean way to prevent them from being loaded
    or to unregister them.
    """
    FormatChecker.check_lines = check_lines



# enable: C0103

class FooBarBaz:
    """
    A good class name.
    """



class Foo:
    """
    A good class name.
    """

class _FooBarZar:
    """
    A class name begins with an underscore.
    """



# enable: C0103

# Function names should be same as method names

def fooBar():
    pass


def foo():
    pass


def _foo():
    pass


def _fooBar():
    pass


def get_fooBar():
    pass


def _SOME_THING():
    pass


def test_SOME_THING():
    pass


# enable: C0103

# Argument names are same as variable names

# Bad examples
# Pylint will generate two warnings for each bad argument name,
# one for invalid argument name, and one for invalid variable name.

def functiona(FooBar):
    pass

def functionb(Foo):
    pass

def functionc(FOO):
    pass

def functiond(foo_bar):
    pass

# Good examples
def functione(f):
    pass

def functionf(foo):
    pass

def functiong(_foo):
    pass

def functionh(fooBar):
    pass



from twisted.trial import unittest

from twistedchecker.core.util import isTestModule, moduleNeedsTests



class UtilTestCase(unittest.TestCase):
    """
    Test for twistedchecker.core.util
    """

    def test_isTestMoudle(self):
        """
        isTestMoudle returns True if the given module name has '.test.' in it.
        """
        self.assertTrue(isTestModule("twisted.test.test_dict"))
        self.assertTrue(isTestModule("twisted.python.test.test_util"))
        self.assertFalse(isTestModule("twisted.test.iosim"))
        self.assertFalse(isTestModule("twisted.python.hook"))


    def test_moduleNeedsTests(self):
        """
        moduleNeedsTests returns true
        if the module name starts with an underscore.
        """
        self.assertFalse(moduleNeedsTests("twisted.python.__init__"))
        self.assertFalse(moduleNeedsTests("twisted.python._release"))
        self.assertTrue(moduleNeedsTests("twisted.python.hook"))



# enable: C9302

class BadMethodNames:
    """
    Methods should be in mixed case, with the first letter lower case,
    each word separated by having its first letter capitalized.
    """

    def CapitalCase(self):
        pass


    def ALLCAPITALCASE(self):
        pass


    def ssh_other_command(self):
        """
        Dispatched command with has no other similar command.

        The checker will consider this an error.
        """
        pass


    def publicMethod_name(self):
        """
        Method names should be fully in camel case.
        """
        pass


    def _privateMethods_not_all_camel(self):
        pass


    def __hiddenMethods_not_all_camel(self):
        pass


    def ___toPrivate(self):
        """
        Method can not start with 3 underscores.
        """

    def __reservedShortEnd_(self):
        pass


    def __reservedLongEnd___(self):
        pass


class SomeInvalidMixin(object):
    """
    Mixin with test methods.
    """

    def testLegacyMethod(self):
        """
        Legacy test method names should no longer be accepted.
        """
        pass


class SomeInvalidTests(object):
    """
    Test suite implementation.
    """

    def testLegacyMethod(self):
        """
        Legacy test method names should no longer be accepted.
        """
        pass


class GoodMethodNames:

    def test(self):
        pass


    def _test(self):
        pass


    def _testConnection(self):
        pass


    def _testPrivateConnection(self):
        pass


    def testPublicConnection(self):
        """
        This method name would have not been valid if this class is a
        test suite.
        """
        pass

    def _FOOBAR(self):
        pass


    def _normalPrivateMethod(self):
        pass


    def __hiddenMethod(self):
        pass


    def __reservedMethod__(self):
        pass


    def __reserved_method__(self):
        pass


    def ftp_COMMAND(self):
        """
        Dispatched command with at least another sibling.

        The checker will heuristically consider this a valid method.
        """
        pass


    def ftp_OTHER_COMMAND(self):
        pass


    def global_port_forwarding(self):
        """
        Dispatched with a sibling with lower case.
        """
        pass


    def global_cancel_port_forwarding(self):
        pass


    def test_SOME_THING(self):
        """
        In Twisted we have methods like render_POST, xmlrpc_SOMETHING and due
        to this test methods like test_SOME_THING or
        test_render_POSTSomething are valid.
        """
        pass


    def test_render_POSTSomething(self):
        pass



# enable: W9201,W9202,W9203,W9204,W9205,W9206,W9207,W9208,W9209
"""
  A docstring with a wrong indentation.
  Docstring should have consistent indentations.
"""

class foo:
    '''The opening/closing of docstring should be on a line by themselves'''

    def a(self):
       """
         A docstring with a wrong indentation.
         """
       pass


    def b(self, argument1):
        """
        This docstring should contain epytext markups of argument1.
        """
        pass


    def c(self, argument1):
        """
        This docstring should contain a @param markup of argument1.

        @type argument1: C{str}
        """
        pass


    def d(self, _argument1):
        """
        This docstring should contain a @type markup of _argument1.

        @param _argument1: a argument
        """
        pass


    def e(self):
        """
        This docstring should contain a @return and @rtype markup.
        """
        return 1


    def f(self):
        """
        This docstring should contain a @return markup.

        @rtype: C{int}
        """
        return 1


    def g(self):
        """
        This docstring should contain a @rtype markup.

        @return: just return 1
        """
        return 1


    def h(self, a):
        """
        There should be a blank line before epytext markups
        @param a: a argument
        @type a: C{int}
        @return: return a
        @rtype: C{int}
        """
        return a


    def i(self):
        '''
            A docstring begins with single quotes,
            and have wrong indentation.
        '''
        pass


    def aCrazyMethod(this, self):
        """
        This is an example method using 'this' as a replacement of 'self'.
        This docstring should contain epytext for the parameter 'self'.
        By the way, this example will cause a fatal error E0213 in pylint.
        """
        pass


    def l(self):
        # Missing docstring.
        pass


    def m(self):
        """
        """
        # Empty docstring.
        pass


    @property
    def goodGetterBadSetter(self):
        """
        Getter properties used to check for invalid setter docstring.
        """
        return True


    @goodGetterBadSetter.setter
    def goodGetterBadSetter(self, val):
        """
        Setter properties docstring exception is only for arguments
        named `value`.
        """


def moduleLevelFunctionWithoutDocstring():
    pass



import os
import sys

from pylint.interfaces import IReporter
from pylint.reporters.text import TextReporter
import twistedchecker


class LimitedReporter(TextReporter):
    """
    A reporter only used by unit tests.
    """
    __implements__ = IReporter
    messagesAllowed = None
    extension = 'txt'
    pathMessageCache = os.path.join(twistedchecker.abspath,
                                    "configuration", "messages.cache")

    def __init__(self, messagesAllowed, output=sys.stdout):
        """
        Initiate the the limited reporter.
        Load a list of allowed messages.

        @param output: output stream
        """
        TextReporter.__init__(self, output)
        self.messagesAllowed = messagesAllowed


    def handle_message(self, msg):
        """
        Manage message of different type and in the context of path.
        """
        if msg.msg_id in self.messagesAllowed:
            super(LimitedReporter, self).handle_message(msg)



# enable: W9701

# A test class name should end with Tests (for instance, FooTests).
# Here we are assuming that a test class is any class that contains
# one or more test_* methods.

from twisted.trial import unittest


class SpecializedTestCaseSubclass(unittest.TestCase):
    """
    A specialized TestCase subclass that test classes can
    inherit from. Note that even though this class inherits
    from TestCase, this class should not be checked by the
    checker, as this class does not contain a test_* method.
    """
    def doSomething(self):
        """
        Some method.
        """
        pass


class SomeTestMixin(object):
    """
    A mixin which does nothing.
    """

    def testLegacyName(self):
        """
        A test with old naming convention.
        """

#
# Bad examples
#
class SomethingTestCase(unittest.TestCase):
    """
    An incorrectly named test class.
    """
    def test_something(self):
        """
        A test method.
        """
        pass



class SomethingElseTest(unittest.TestCase, SomeTestMixin):
    """
    Another incorrectly named test class.
    """
    def test_somethingElse(self):
        """
        A test method.
        """
        pass



class TestFoo(SpecializedTestCaseSubclass):
    """
    One more incorrectly named test class.
    """
    def test_foo(self):
        """
        A test method.
        """
        pass



class LegacyTestCase(unittest.TestCase):
    """
    An incorrectly named test class.
    """
    def testSomething(self):
        """
        A test method with old naming convention.
        """


#
# Good examples
#
class SampleTestMixin(object):
    """
    A sample mixin with additional assertion helpers.
    """
    def assertSomething(self):
        """
        An assertion helper.
        """
        pass

    def test_someTest(self):
        """
        Test from mixin.
        """
        pass



class SomethingTests(unittest.TestCase):
    """
    A correctly named test class.
    """
    def test_something(self):
        """
        A test method.
        """



class SomethingElseTests(unittest.TestCase, SampleTestMixin):
    """
    Another correctly named test class.
    """
    def test_somethingElse(self):
        """
        A test method.
        """



class FooTests(SpecializedTestCaseSubclass):
    """
    One more correctly named test class.
    """
    def test_foo(self):
        """
        A test method.
        """



class LegacyTests(SpecializedTestCaseSubclass):
    """
    One more correctly named with test methods after older convention.
    """
    def testLegacyName(self):
        """
        A test method.
        """



# -*- test-case-name: twistedchecker.test.test_runner -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

import sys
import os
import re

from pylint.checkers.base import NameChecker
from pylint.lint import PyLinter
from logilab.common.modutils import file_from_modpath

from twisted.python.compat import NativeStringIO

import twistedchecker
from twistedchecker.checkers import patch_pylint_format
from twistedchecker.core.exceptionfinder import findAllExceptions
from twistedchecker.reporters.limited import LimitedReporter


class Runner():
    """
    Run and control the checking process.
    """
    outputStream = None
    linter = None
    allowOptions = None
    # Customized checkers.
    checkers = ("header.HeaderChecker",
                "names.TwistedNamesChecker",
                "docstring.DocstringChecker",
                "formattingoperation.FormattingOperationChecker",
                "comment.CommentChecker",
                "testclassname.TestClassNameChecker")
    allowedMessagesFromPylint = ("F0001",
                                 "C0103",
                                 "C0301",
                                 "W0311",
                                 "W0312")
    diffOption = None
    errorResultRead = "Error: Failed to read result file '%s'.\n"
    prefixModuleName = "************* Module "
    regexLineStart = "^[WCEFR]\d{4}\:"

    def __init__(self):
        """
        Initialize C{PyLinter} object, and load configuration file.
        """
        self.allowOptions = True
        self.linter = PyLinter(self._makeOptions())
        # register standard checkers.
        self.linter.load_default_plugins()
        # read configuration.
        pathConfig = os.path.join(twistedchecker.abspath,
                                  "configuration", "pylintrc")
        self.linter.read_config_file(pathConfig)
        # now we can load file config and command line, plugins (which can
        # provide options) have been registered.
        self.linter.load_config_file()
        allowedMessages = self.registerCheckers()
        # disable messages
        disabledMessages = set(self.linter
                           .cfgfile_parser.get("TWISTEDCHECKER", "disable")
                           .replace(" ", "").split(","))
        if disabledMessages != {""}:
            for msg in disabledMessages:
                self.linter.disable(msg)
            allowedMessages -= disabledMessages
        # set default output stream to stdout
        self.setOutput(sys.stdout)
        # set default reporter to limited reporter
        self.setReporter(LimitedReporter(allowedMessages))


    def _makeOptions(self):
        """
        Return options for twistedchecker.
        """
        return (
            ("diff",
             {"type": "string",
              "metavar": "<result-file>",
              "help": "Set comparing result file to automatically "
                      "generate a diff."}
            ),
            ('pep8',
             {'type': 'yn', 'metavar': '<y_or_n>',
              'default': False,
              'help': 'Show pep8 warnings.'}
            ),
            ('strict-epydoc',
             {'type': 'yn', 'metavar': '<y_or_n>',
              'default': False,
              'help': "Check '@type' and '@rtype' in epydoc."}
            ),
          )


    def setOutput(self, stream):
        """
        Set the stream to output result of checking.

        @param stream: output stream, defaultly it should be stdout
        """
        self.outputStream = stream
        sys.stdout = stream


    def setReporter(self, reporter):
        """
        Set the reporter of pylint.

        @param reporter: reporter used to show messages
        """
        self.linter.set_reporter(reporter)


    def displayHelp(self):
        """
        Output help message of twistedchecker.
        """
        self.outputStream.write(self.linter.help())
        sys.exit(32)


    def registerCheckers(self):
        """
        Register all checkers of TwistedChecker to C{PyLinter}.

        @return: a list of allowed messages
        """
        # We patch the default pylint format checker.
        patch_pylint_format.patch()

        # register checkers
        allowedMessages = list(self.allowedMessagesFromPylint)
        for strChecker in self.checkers:
            modname, classname = strChecker.split(".")
            strModule = "twistedchecker.checkers.%s" % modname
            checker = getattr(__import__(strModule,
                                        fromlist=["twistedchecker.checkers"]),
                             classname)
            instanceChecker = checker(self.linter)
            allowedMessages += list(instanceChecker.msgs.keys())
            self.linter.register_checker(instanceChecker)

        self.restrictCheckers(allowedMessages)
        return set(allowedMessages)


    def unregisterChecker(self, checker):
        """
        Remove a checker from the list of registered checkers.

        @param checker: the checker to remove
        """
        self.linter._checkers[checker.name].remove(checker)
        if checker in self.linter._reports:
            del self.linter._reports[checker]
        if checker in self.linter.options_providers:
            self.linter.options_providers.remove(checker)


    def findUselessCheckers(self, allowedMessages):
        """
        Find checkers which generate no allowed messages.

        @param allowedMessages: allowed messages
        @return: useless checkers, remove them from pylint
        """
        uselessCheckers = []
        for checkerName in self.linter._checkers:
            for checker in list(self.linter._checkers[checkerName]):
                messagesOfChecker = set(checker.msgs)
                if not messagesOfChecker.intersection(allowedMessages):
                    uselessCheckers.append(checker)
        return uselessCheckers


    def restrictCheckers(self, allowedMessages):
        """
        Unregister useless checkers to speed up twistedchecker.

        @param allowedMessages: output messages allowed in twistedchecker
        """
        uselessCheckers = self.findUselessCheckers(allowedMessages)
        # Unregister these checkers
        for checker in uselessCheckers:
            self.unregisterChecker(checker)


    def getCheckerByName(self, checkerType):
        """
        Get checker by given name.

        @checkerType: type of the checker
        """
        for checker in sum(list(self.linter._checkers.values()), []):
            if isinstance(checker, checkerType):
                return checker
        return None


    def allowPatternsForNameChecking(self, patternsFunc, patternsClass):
        """
        Allow name exceptions by given patterns.

        @param patternsFunc: patterns of special function names
        @param patternsClass: patterns of special class names
        """
        cfgParser = self.linter.cfgfile_parser
        nameChecker = self.getCheckerByName(NameChecker)
        if not nameChecker:
            return
        if patternsFunc:
            regexFuncAdd = "|((%s).+)$" % "|".join(patternsFunc)
        else:
            regexFuncAdd = ""
        if patternsClass:
            regexClassAdd = "|((%s).+)$" % "|".join(patternsClass)
        else:
            regexClassAdd = ""
        # Modify regex for function, method and class name.
        regexMethod = cfgParser.get("BASIC", "method-rgx") + regexFuncAdd
        regexFunction = cfgParser.get("BASIC", "function-rgx") + regexFuncAdd
        regexClass = cfgParser.get("BASIC", "class-rgx") + regexClassAdd
        # Save to config parser.
        cfgParser.set("BASIC", "method-rgx", regexMethod)
        cfgParser.set("BASIC", "function-rgx", regexFunction)
        cfgParser.set("BASIC", "class-rgx", regexClass)
        # Save to name checker.
        nameChecker.config.method_rgx = re.compile(regexMethod)
        nameChecker.config.function_rgx = re.compile(regexFunction)
        nameChecker.config.class_rgx = re.compile(regexClass)


    def getPathList(self, filesOrModules):
        """
        Transform a list of modules to path.

        @param filesOrModules: a list of modules (may be foo/bar.py or
        foo.bar)
        """
        pathList = []
        for fileOrMod in filesOrModules:
            if not os.path.exists(fileOrMod):
                # May be given module is not not a path,
                # then transform it to a path.
                try:
                    filepath = file_from_modpath(fileOrMod.split('.'))
                except (ImportError, SyntaxError):
                    # Could not load this module.
                    continue
                if not os.path.exists(filepath):
                    # Could not find this module in file system.
                    continue
                if os.path.basename(filepath) == "__init__.py":
                    filepath = os.path.dirname(filepath)
            else:
                filepath = fileOrMod
            pathList.append(filepath)
        return pathList


    def setNameExceptions(self, filesOrModules):
        """
        Find name exceptions in codes and allow them to be ignored
        in checking.

        @param filesOrModules: a list of modules (may be foo/bar.py or
        foo.bar)
        """
        pathList = self.getPathList(filesOrModules)
        for path in pathList:
            patternsFunc, patternsClass = findAllExceptions(path)
            self.allowPatternsForNameChecking(patternsFunc, patternsClass)


    def run(self, args):
        """
        Setup the environment, and run pylint.

        @param args: arguments will be passed to pylint
        @type args: list of string
        """
        # set output stream.
        if self.outputStream:
            self.linter.reporter.set_output(self.outputStream)
        try:
            args = self.linter.load_command_line_configuration(args)
        except SystemExit as exc:
            if exc.code == 2:  # bad options
                exc.code = 32
            raise
        if not args:
            self.displayHelp()
        # Check for 'strict-epydoc' option.
        if self.allowOptions and not self.linter.option_value("strict-epydoc"):
            for msg in ["W9203", "W9205"]:
                self.linter.disable(msg)

        # insert current working directory to the python path to have a correct
        # behaviour.
        sys.path.insert(0, os.getcwd())
        # set exceptions for name checking.
        self.setNameExceptions(args)

        # check for diff option.
        self.diffOption = self.linter.option_value("diff")
        if self.diffOption:
            self.prepareDiff()

        # check codes.
        self.linter.check(args)

        # show diff of warnings if diff option on.
        if self.diffOption:
            diffCount = self.showDiffResults()
            exitCode = 1 if diffCount else 0
            sys.exit(exitCode)

        sys.exit(self.linter.msg_status)


    def prepareDiff(self):
        """
        Prepare to run the checker and get diff results.
        """
        self.streamForDiff = NativeStringIO()
        self.linter.reporter.set_output(self.streamForDiff)


    def showDiffResults(self):
        """
        Show results when diff option on.
        """
        try:
            oldWarnings = self.parseWarnings(self._readDiffFile())
        except:
            sys.stderr.write(self.errorResultRead % self.diffOption)
            return 1

        newWarnings = self.parseWarnings(self.streamForDiff.getvalue())

        diffWarnings = self.generateDiff(oldWarnings, newWarnings)

        if diffWarnings:
            diffResult = self.formatWarnings(diffWarnings)
            self.outputStream.write(diffResult + "\n")
            return len(diffWarnings)
        else:
            return 0

    def _readDiffFile(self):
        """
        Read content of diff file.

        This is here to help with testing.

        @return: File content.
        @rtype: c{str}
        """
        with open(self.diffOption) as f:
            content = f.read()
        return content

    def generateDiff(self, oldWarnings, newWarnings):
        """
        Generate diff between given two lists of warnings.

        @param oldWarnings: parsed old warnings
        @param newWarnings: parsed new warnings
        @return: a dict object of diff
        """
        diffWarnings = {}

        for modulename in newWarnings:
            diffInModule = (
                newWarnings[modulename] -
                oldWarnings.get(modulename, set()))
            if diffInModule:
                diffWarnings[modulename] = diffInModule

        return diffWarnings


    def parseWarnings(self, result):
        """
        Transform result in string to a dict object.

        @param result: a list of warnings in string
        @return: a dict of warnings
        """
        warnings = {}
        currentModule = None
        warningsCurrentModule = []
        for line in result.splitlines():
            if line.startswith(self.prefixModuleName):
                # Save results for previous module
                if currentModule:
                    warnings[currentModule] = set(warningsCurrentModule)
                # Initial results for current module
                moduleName = line.replace(self.prefixModuleName, "")
                currentModule = moduleName
                warningsCurrentModule = []
            elif re.search(self.regexLineStart, line):
                warningsCurrentModule.append(line)
            else:
                if warningsCurrentModule:
                    warningsCurrentModule[-1] += "\n" + line
        # Save warnings for last module
        if currentModule:
            warnings[currentModule] = set(warningsCurrentModule)
        return warnings


    def formatWarnings(self, warnings):
        """
        Format warnings to a list of results.

        @param warnings: a dict of warnings produced by parseWarnings
        @return: a list of warnings in string
        """
        lines = []
        for modulename in sorted(warnings):
            lines.append(self.prefixModuleName + modulename)
            lines.extend(sorted(warnings[modulename],
                         key=lambda x: x.split(":")[1]))

        return "\n".join(lines)


def main():
    """
    An entry point used in the setup.py to create a runnable script.
    """
    runner = Runner()
    runner.run(sys.argv[1:])



"""
Utility functions for twistedchecker.
"""

def isTestModule(modulename):
    """
    Determine whether a module is a test module
    by given module name.
    """
    filename = modulename.split('.')[-1]
    return ".test." in modulename and filename.startswith('test')



def moduleNeedsTests(modulename):
    """
    Determine whether a module is a is a special module.
    like __init__
    """
    return not modulename.split(".")[-1].startswith("_")



__all__ = ["isTestModule", "moduleNeedsTests"]



# enable: C9302,C9303

class SomeInvalidMixin(object):
    """
    Mixin with test methods.
    """

    def testLegacyMethod(self):
        """
        Legacy test method names should no longer be accepted.
        """
        pass


class SomeInvalidTests(object):
    """
    Test suite implementation.
    """

    def testLegacyMethod(self):
        """
        Legacy test method names should no longer be accepted.
        """
        pass


class SomeValidTests(object):
    """
    This test suite class contains only valid method named.
    """

    def _normalPrivateMethod(self):
        pass


    def normalMethod(self):
        pass


    def test_SOME_THING(self):
        """
        In Twisted we have methods like render_POST, xmlrpc_SOMETHING and due
        to this test methods like test_SOME_THING or
        test_render_POSTSomething are valid.
        """
        pass


    def test_render_POSTSomething(self):
        pass



# enable: W9301

# A test module begins with test_

from twisted.trial import unittest


class SubclassOfTestCase(unittest.TestCase):
    """
    A subclass of TestCase.
    """
    def test_something(self):
        """
        A test method.
        """
        pass



"""
Checker for naming convention.
"""
import re

from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker

from twistedchecker.core.util import isTestModule


class TwistedNamesChecker(BaseChecker):
    """
    A checker for checking Twisted naming convention.
    """
    msgs = {
        'W9301': ('Test modules should begin with test_',
                  'Used when a test module begins with test_.',
                  'testmodule-not-beginning-with-test'),
        'C9302': ('Method name is invalid',
                  'Used when a method has invalid name.',
                  'invalid-method-name'),
        'C9303': ('Test method name is invalid',
                  'Used when a test method has invalid name.',
                  'invalid-test-method-name'),
    }
    __implements__ = IAstroidChecker
    name = 'modulename'
    options = ()

    def moduleContainsTestCase(self, node):
        """
        Determine whether a module contains a subclass of TestCase.

        @param node: node of given module
        """
        patternTestCase = br"class\s+[a-zA-Z0-9]+\s*\(.*TestCase\)"
        return re.search(patternTestCase, node.file_stream.read()) \
               and True or False


    def visit_module(self, node):
        """
        A interface will be called when visiting a module.

        @param node: node of current module
        """
        modulename = node.name.split(".")[-1]
        if isTestModule(node.name) and self.moduleContainsTestCase(node):
            self._checkTestModuleName(modulename, node)


    def visit_functiondef(self, node):
        """
        A interface will be called when visiting a function or a method.

        @param node: the current node
        """
        if not node.is_method():
            # We only check methods.
            return

        name = node.name

        if isTestModule(node.root().name):
            if name.startswith('test'):
                if not name.startswith('test_'):
                    self.add_message('C9303', node=node)
                    return
                else:
                    # Test names start with 'test_NAME' and can be like
                    # test_SOME_NAME or test_render_SomeCondition.
                    return

        if name[0].isupper():
            self.add_message('C9302', node=node)
            return

        if name.startswith('___'):
            self.add_message('C9302', node=node)
            return

        if name.endswith('___'):
            self.add_message('C9302', node=node)
            return

        if name.startswith('__'):
            if name.endswith('___'):
                # To many trailing underscores.
                self.add_message('C9302', node=node)
                return
            if name.endswith('_') and not name.endswith('__'):
                # To few trailing underscored
                self.add_message('C9302', node=node)
                return
            if name.endswith('__'):
                # This is a reserved name and we don't do any checks on it.
                return
            name = name[2:-2]

        if name.startswith('_'):
            name = name[1:]

        if name.endswith('_'):
            self.add_message('C9302', node=node)
            return

        if '_' in name:
            # This has a underscore in the main name.
            prefix = self._getMethodNamePrefix(node)
            if prefix:
                # There are other names with same prefix so this should be
                # a dispatched method.
                return

            self.add_message('C9302', node=node)
            return


        if isTestModule(node.name) and self.moduleContainsTestCase(node):
            self._checkTestMethodName(node)

    def _getMethodNamePrefix(self, node):
        """
        Return the prefix of this method based on sibling methods.

        @param node: the current node
        """
        targetName = node.name
        for sibling in node.parent.nodes_of_class(type(node)):
            if sibling is node:
                # We are on the same node in parent so we skip it.
                continue
            prefix = self._getCommonStart(targetName, sibling.name)
            if not prefix.rstrip('_'):
                # We ignore prefixes which are just underscores.
                continue
            return prefix

        return ''


    def _getCommonStart(self, left, right):
        """
        Return the common prefix of the 2 strings.

        @param left: one string
        @param right: another string
        """
        prefix = []
        for a, b in zip(left, right):
            if a == b:
                prefix.append(a)
            else:
                break

        return ''.join(prefix)


    def _checkTestModuleName(self, modulename, node):
        """
        Check whether a test module has correct module name.
        The module name should begins with test_.

        @param modulename: a module name
        """
        if not modulename.startswith("test_"):
            self.add_message('W9301', node=node)


    def _checkTestMethodName(self, modulename, node):
        """
        Check whether a test method has correct module name.

        @param modulename: a module name
        """
        if not modulename.startswith("test_"):
            self.add_message('W9301', node=node)



# enable: C0103

MODULELEVELCONSTANT = True

class FooClass:
    """
    Just a class.
    """
    def __init__(self):
        """
        Some bad variable names.
        """
        f = None
        foo = None
        _foo = None
        fooBar = None



import re
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker

COMMENT_RGX = re.compile(b"#.*$", re.M)

class CommentChecker(BaseChecker):
    """
    A checker for checking comment issues.

    A good comment should begin with one whitespace and
    with first letter capitalized.
    """
    msgs = {
     'W9401': ('Comments should begin with one whitespace',
               'Used for checking comment format issues.',
                   'comments-one-whitespace'),
     'W9402': ('The first letter of comment should be capitalized',
               'Used for checking comment format issues. ',
                   'comments-capitalized')
    }
    __implements__ = IAstroidChecker
    name = 'comment'
    options = ()

    def visit_module(self, node):
        """
        A interface will be called when visiting a module.

        @param node: node of current module
        """
        if not node.file_stream:
            # Failed to open the module
            return
        isFirstLineOfComment = True
        isDocString = False
        lines = node.stream().readlines()
        for linenum, line in enumerate(lines):
            if line.strip().startswith(b'"""'):
                # This is a simple assumption than docstring are delimited
                # with triple double quotes on a single line.
                # Should do the job for Twisted code.
                isDocString = not isDocString

            if isDocString:
                # We ignore comments in docstrings.
                continue

            matchedComment = COMMENT_RGX.search(line)
            if matchedComment:
                if isFirstLineOfComment:
                    # Check for W9401
                    comment = matchedComment.group()
                    if (comment.startswith(b"#  ") or
                        not comment.startswith(b"# ")):
                        self.add_message('W9401', line=linenum + 1, node=node)
                    # Check for W9402
                    strippedComment = comment.lstrip(b"#").lstrip()
                    if strippedComment:
                        firstLetter = strippedComment[0:1]
                        if (firstLetter.isalpha() and
                            not firstLetter.isupper()):
                            self.add_message('W9402', line=linenum + 1, node=node)
                    isFirstLineOfComment = False
            else:
                isFirstLineOfComment = True



# enable: W9501
# -*- test-case-name: twistedchecker.test.test_functionaltests -*-

num = 3
# we should use a tuple as the value
formattedString = "%d" % num

formattedString = "%d" % (num)

# a formatting operation using mapping value
# no warnings should be generated
mapFoo = {'num': 3}
formattedString = "%(num)d" % mapFoo

num = 3
# a tuple used in the string formatting operation.
formattedString = "%d" % (num,)

# a formatting operation using mapping value
# no warnings should be generated
constantFormat = "a format %(value)s"
constantFormat % {"value": "value"}



# enable: C0103

class FooClass:
    """
    Attribute names should be in mixed case,
    with the first letter lower case,
    each word separated by having its first letter capitalized
    just like method names.
    And all private names should begin with an underscore.
    """

    def __init__(self):
        """
        A init of this class
        """
        # Bad names.
        self.foo_bar_zar = None
        self.FooBarZar = None
        self.a = None
        self._fooBar_ = None
        self._fooBar__ = None
        self.__fooBar_ = None
        self.___longStart__ = None
        self.__longEnd___ = None

        # Good names.
        self.fooBarZar = None
        self._fooBarZar = None
        self.foobar = None
        # Any __*__ name is accepted.
        self.__foobar__ = None
        self.__fooBar__ = None
        self.__foo_bar__ = None
        self.__dict__ = {}
        self.__version__ = {}
        self.__a__ = {}
        # In some cases constants are lazy initialized as instance members.
        self.FOOBARZAR = None



"""
TwistedChecker

A project of GSoC 2012 to checker for codes of twisted.
"""

import os
import sys

# set the absolute path of TwistedChecker
abspath = os.path.dirname(os.path.abspath(__file__))

version = getattr(sys, "version_info", (0,))
if version < (2, 7):
    raise ImportError("TwistedChecker requires Python 2.7 or Python 3.5+ ")
elif version >= (3, 0) and version < (3, 5):
    raise ImportError("TwistedChecker on Python 3 requires Python 3.5+")



from pylint.interfaces import IAstroidChecker
from pylint.checkers.strings import StringFormatChecker



class FormattingOperationChecker(StringFormatChecker):
    """
    When string formatting operations are used like formatString % values,
    we should always use a tuple for non-mapping values.
    """
    msgs = {
     'W9501': ('String formatting operations should always use a tuple '
               'for non-mapping values',
               'Checking string formatting operations.', 'formatting-use-tuple'),
    }
    __implements__ = IAstroidChecker
    name = 'formattingoperation'
    options = ()

    def visit_binop(self, node):
        """
        Called when if a binary operation is found.
        Only check for string formatting operations.

        @param node: currently checking node
        """
        if node.op != "%":
            return
        pattern = node.left.as_string()
        # If the pattern's not a constant string, we don't know whether a
        # dictionary or a tuple makes sense, so don't try to guess.
        if not pattern.startswith("'") or pattern.startswith('"'):
            return
        # If the pattern has things like %(foo)s, then the values can't be a
        # tuple, so don't check for it.
        if "%(" in pattern:
            return
        valueString = node.right.as_string()
        tupleUsed = valueString.startswith('(')
        if tupleUsed:
            return
        self.add_message('W9501', node=node)



# enable: W9201,W9202,W9203,W9204,W9205,W9206,W9207,W9208,W9209
# -*- test-case-name: twistedchecker.test.test_functionaltests.FunctionalTests.test_twistedchecker_functionaltests_docstring_pass -*-
"""
A docstring with a wrong indentation.
Docstring should have consistent indentations.
"""

from elsewhere import aliasForProperty


class SomeOptions:
    """
    Certain behavior differs for CLI option implementations.
    """
    def opt_bar(self, param):
        """
        An option docstring need not have param markup.
        """



class foo:
    """
    The opening/closing of docstring should be on a line by themselves.
    """

    def a(self):
        """
        A docstring with a correct indentation.
        """
        pass


    def b(self, argument1):
        """
        This docstring should contain epytext markups of argument1.

        @param argument1: an argument
        @type argument1: C{int}
        """
        pass


    def c(self):
        """
        This docstring should contain a @return and @rtype markup.

        @return: return 1
        @rtype: C{int}
        """
        return 1


    def d(self, a):
        """
        There should be a blank line before epytext markups.

        @param a: a argument
        @type a: C{int}
        @return: return a
        @rtype: C{int}
        """
        return a


    def e(self):
        """
        A method contains a function k().
        No warning should be generated
        """
        def k():
            pass


    @property
    def someProperty(self):
        """
        Getter properties don't need return docstring.
        """
        return True


    @someProperty.setter
    def someProperty(self, value):
        """
        Setter properties don't need to document their value.
        """

    @property
    def otherProperty(self):
        """
        Getter properties used the test that setter don't need a docstring.
        """
        return True


    @otherProperty.setter
    def otherProperty(self, value):
        # Setter don't need a docstring as most of the time is a duplicate of
        # the getter docstring.
        pass


    @aliasForProperty.setter
    def decorated(self):
        # If we have a decorator which *might* be a property, but is imported
        # from elsewhere (somewhere that pylint's parser can't see) then we
        # also don't need a docstring for its setter.
        pass


    def f(self):
        """
        A method returns nothing.
        """
        return


    def test_someTest(self):
        """
        Test methods in trial should not document their return value.
        """
        return 'deferred'



class Bar(object):
    """
    A cvar is recognized as being the start of epytext markup.

    @cvar foo: bar baz
    @type foo: bax
    """

    def a(self):
        """
        A raises field is recognized as being the start of epytext markup.

        @raises FoobarException: An exception.
        @returns: C{int}
        """

    def b(self):
        """
        A raise (because we are inconsistent about stuff) should also be
        recognized as the start of epytext markup.

        @raise BarException: Another exception.
        @returns: C{int}
        """
        def callback(result):
            pass



def topLevel():
    """
    A top-level function.
    """
    class Inner(object):
        def innerInner(self):
            pass



class Baz(object):
    """
    An ivar is recognized as being the start of epytext markup.

    @ivar foo: bar baz
    @type foo: bax
    """



# enable: C0103

class fooBarBaz:
    """
    Class name begins with lower-case letter.
    """



class boo:
    """
    Class name begins with lower-case letter.
    """



class Foo_Bar_Baz:
    """
    Class name contains underscore.
    """



# enable: C0301
# max line length of Twsited projects should be 79
print("this line is long long  long  long long long long long ends at column 81")

# But comments with url should be excepted.
# http://thisurlisverylongandwillneverfitinto80columnsnomatterwhat.example.com/somegreatcontent.tml

""""
Long url form docstrings are ignored.
See U{https://thisurlisverylongandwillneverfitinto80columnsnomatterwhat.example.com/somegreatcontent.html}
"""



#!/usr/bin/env python
# -*- test-case-name: twistedchecker -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

from setuptools import find_packages, setup

with open('README.rst') as f:
    longDescription = f.read()

setup(
    name='twistedchecker',
    description='A Twisted coding standard compliance checker.',
    version='0.7.3',
    author='Twisted Matrix Laboratories',
    author_email='twisted-python@twistedmatrix.com',
    url='https://github.com/twisted/twistedchecker',
    packages=find_packages(),
    include_package_data=True,  # use MANIFEST.in during install
    entry_points={
      "console_scripts": [
          "twistedchecker = twistedchecker.core.runner:main"
      ]
    },
    license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Quality Assurance"
    ],
    keywords=[
        "twisted", "checker", "compliance", "pep8"
    ],
    install_requires=[
        "pylint == 1.5.6",
        "logilab-common == 1.2.1",
        'twisted>=15.0.0',
    ],
    extras_require = {
        'dev':  [
            'pyflakes',
            'coverage'
            ],
    },
    long_description=longDescription
)



import sys

from pylint.interfaces import IReporter
from pylint.reporters import BaseReporter



class TestReporter(BaseReporter):
    """
    A reporter only used by unit tests.
    """
    __implements__ = IReporter
    extension = 'txt'


    def __init__(self, output=sys.stdout):
        """
        Initiate the the base reporter.

        @param output: output stream
        """
        BaseReporter.__init__(self, output)


    def add_message(self, msg_id, location, msg):
        """
        Manage message of different type and in the context of path.
        Output error message in format of [line number]: [message id]

        @param msg_id: message id
        @param location: detailed location information
        @param msg: text add_message
        """
        module, obj, line, col_offset = location[1:]
        self.writeln('%s:%s' % (line, msg_id))



import operator

from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker
from astroid.scoped_nodes import ClassDef

from twistedchecker.core.util import isTestModule


class TestClassNameChecker(BaseChecker):
    """
    A checker for checking test class names.

    Test classes should be named FooTests, where Foo is the name
    of the component/feature being tested.
    """
    msgs = {
        'W9701': ('Test class names should end with Tests',
                  'Used for checking test class names.',
                  'test-class-end-with-Tests'),
    }
    __implements__ = IAstroidChecker
    name = 'testclassname'
    options = ()

    def visit_module(self, node):
        """
        Called when the AST checker visits a module.

        @param node: node of current module
        """
        if not isTestModule(node.name):
            return
        objects = list(node.values())
        objects.sort(key=operator.attrgetter('lineno'))
        for obj in objects:
            if (isinstance(obj, ClassDef) and self._isTestClass(obj) and
                    not obj.name.endswith('Tests')):
                self.add_message('W9701', line=obj.lineno, node=obj)


    def _isTestClass(self, klass):
        """
        Check whether a class is a test class. Here we assume that a
        test class is any class that contains one or more test_* or
        test* methods and inherits from TestCase.

        @param klass: Class to be checked for whether it is a test class.
        @type klass: L{logilab.astng.scoped_nodes.Class}

        @return: C{True} if the given class is a test class, C{False}
            otherwise.
        @rtype: L{bool}
        """
        ancestors = [ancestor.name for ancestor in klass.ancestors()]
        methods = [method.name for method in klass.mymethods()]

        if 'TestCase' not in ancestors:
            return False

        for method in methods:
            if method.startswith('test'):
                return True
        return False



# enable: C0103

# Bad examples

foo_bar = None

foo_barBaz = None

foo_Bar = None

# Good examples

FOO = None

FOO_BAR = None

Foo = None

FooBar = None

__FOO__ = None

foo = None

fooBar = None

_fooBar = None


class SomeClass(object):

    SOME_CONSTANT = 'some_value'

    def __init__(self):
        self.variable = self.SOME_CONSTANT
        # Constants can also act as instance variables,
        # for example, late initialization.
        self.SOME_CONSTANT = 'new-value'

    def otherMethod(me):
        me.variable = self.SOME_CONSTANT
        # Constants can also act as instance variables,
        # for example, late initialization.
        me.SOME_CONSTANT = 'new-value'

    @classmethod
    def someMethod(cls):
        cls.variable = cls.SOME_CONSTANT
        # Constants can also act as class variables,
        # for example, late initialization.
        cls.SOME_CONSTANT = 'new-value'



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for L{twistedchecker.core.runner}.
"""

import sys
import os
import operator

from functools import reduce

from twisted.python.compat import NativeStringIO
from twisted.trial import unittest

import twistedchecker
from twistedchecker.core.runner import Runner
from twistedchecker.checkers.header import HeaderChecker
from twistedchecker.reporters.test import TestReporter

from twistedchecker.test.test_exceptionfinder import (
    createTestFiles as createTestFilesForFindingExceptions)



class RunnerTestCase(unittest.TestCase):
    """
    Test for twistedchecker.core.runner.Runner.
    """
    debug = False

    def setUp(self):
        """
        Redirect stdout to a temp C{StringIO} stream.
        """
        self.outputStream = NativeStringIO()
        self.patch(sys, "stdout", self.outputStream)
        self.errorStream = NativeStringIO()
        self.patch(sys, "stderr", self.errorStream)


    def clearOutputStream(self):
        """
        A function to clear output stream.
        """
        self.outputStream = NativeStringIO()


    def makeRunner(self):
        """
        Return a runner instance.
        """
        runner = Runner()
        runner.setOutput(self.outputStream)
        return runner


    def _loadAllowedMessages(self):
        """
        Load allowed messages from test files.
        """
        pathTests = os.path.join(twistedchecker.abspath, "functionaltests")
        testfiles = reduce(operator.add,
                           [[os.path.join(pathDir, f) for f in files if f.endswith(".py")]
                            for pathDir, _, files in os.walk(pathTests)])
        messagesAllowed = set(Runner.allowedMessagesFromPylint)
        for testfile in testfiles:
            with open(testfile) as f:
                firstline = f.readline().strip()
            if (firstline.startswith("#") and "enable" in firstline
                                          and ":" in firstline):
                messages = firstline.split(":")[1].strip().split(",")
                messagesAllowed.update(messages)
        return messagesAllowed


    def test_findUselessCheckers(self):
        """
        Test for method findUselessCheckers
        """
        runner = Runner()
        registeredCheckers = sum(list(runner.linter._checkers.values()), [])
        # remove checkers other than header checker
        headerCheckerList = [ckr
                             for ckr in registeredCheckers
                             if type(ckr) == HeaderChecker]
        self.assertTrue(headerCheckerList)
        headerChecker = headerCheckerList[0]
        uselessCheckers = runner.findUselessCheckers(
                            list(headerChecker.msgs.keys())[:1])
        self.assertEqual(len(uselessCheckers) + 1, len(registeredCheckers))
        self.assertTrue(headerChecker not in uselessCheckers)


    def test_unregisterChecker(self):
        """
        Test for method unregisterChecker.

        Remove HeaderChecker from registered,
        and make sure it was removed.
        """
        runner = Runner()
        registeredCheckers = sum(list(runner.linter._checkers.values()), [])
        # Make sure an instance of HeaderChecker in registered checkers
        headerCheckerList = [ckr
                             for ckr in registeredCheckers
                             if type(ckr) == HeaderChecker]
        self.assertTrue(headerCheckerList)
        headerChecker = headerCheckerList[0]
        # Make sure it in option providers
        self.assertTrue(headerChecker in runner.linter.options_providers)
        runner.unregisterChecker(headerChecker)
        # Make sure the instance of HeaderChecker was removed
        registeredCheckers = sum(list(runner.linter._checkers.values()), [])
        self.assertFalse(headerChecker in registeredCheckers)
        # Could not check reports because HeaderChecker is not be
        # recorded in that list
        # Make sure it was removed from option providers
        self.assertFalse(headerChecker in runner.linter.options_providers)


    def test_restrictCheckers(self):
        """
        Test for method restrictCheckers.

        Manually set allowed messages,
        then check for the result of registered checkers
        after run this method.
        """
        runner = Runner()
        runner.restrictCheckers(list(HeaderChecker.msgs.keys())[:1])
        # After run it, only HeaderChecker should be left in
        # registered checkers
        registeredCheckers = sum(list(runner.linter._checkers.values()), [])
        self.assertEqual(len(registeredCheckers), 1)
        self.assertEqual(type(registeredCheckers[0]), HeaderChecker)


    def test_allMessagesAreRegistered(self):
        """
        A test to assume all tests are registered to reporter.
        """
        linter = Runner().linter
        messagesFromTests = self._loadAllowedMessages()
        messagesFromReporter = linter.reporter.messagesAllowed
        messagesDisabled = set(linter
                           .cfgfile_parser.get("TWISTEDCHECKER", "disable")
                           .replace(" ", "").split(","))
        self.assertEqual(messagesFromTests - messagesDisabled,
                         messagesFromReporter)


    def test_runVersion(self):
        """
        Pass argument "--version" to C{runner.run}, and it should show
        a version infomation, then exit. So that I could know it called pylint.
        """
        runner = Runner()
        runner.setOutput(self.outputStream)

        exitResult = self.assertRaises(SystemExit, runner.run, ["--version"])

        self.assertTrue(self.outputStream.getvalue().count("Python") > 0, \
                        msg="failed to call pylint")
        self.assertIsNone(runner.diffOption)
        self.assertEqual(0, exitResult.code)


    def test_runNoError(self):
        """
        When checked file is clean and has no errors it exit with code 0
        without any other output.
        """
        runner = Runner()
        runner.setOutput(self.outputStream)

        # The twistedchecker/checkers/__init__.py is assumed to be clean.
        exitResult = self.assertRaises(SystemExit, runner.run, [
            "twistedchecker.checkers.__init__"])

        self.assertEqual('', self.outputStream.getvalue())
        self.assertEqual(0, exitResult.code)


    def test_runWithErrors(self):
        """
        When checked file is not clean it will exit with non zero exit code.
        """
        runner = Runner()
        runner.setOutput(self.outputStream)

        # The comments functional test is assumed to have at lest one error.
        exitResult = self.assertRaises(SystemExit, runner.run, [
            "twistedchecker.functionaltests.comments"])

        self.assertNotEqual(0, exitResult.code)


    def test_parseWarnings(self):
        """
        Test for twistedchecker.core.runner.Runner.parseWarnings.
        """
        textWarnings = """
************* Module foo
W9001:  1,0: Missing copyright header
************* Module bar
W9002:  1,0: Missing a reference to test module in header
C0111:  10,0: Missing docstring
        """.strip()

        warningsCorrect = {
            "foo": {"W9001:  1,0: Missing copyright header", },
            "bar": {"W9002:  1,0: Missing a reference "
                    "to test module in header",
                    "C0111:  10,0: Missing docstring"
                   }
        }

        warnings = Runner().parseWarnings(textWarnings)
        self.assertEqual(warnings, warningsCorrect)

    def test_runDiffNoWarnings(self):
        """
        When running in diff mode set path to result file and exit with 0 if
        no warnings were found.
        """
        runner = self.makeRunner()
        # Mock showDiffResults to check that it is called.
        showDiffResultsCalls = []
        runner.showDiffResults = lambda: showDiffResultsCalls.append(True)

        exitResult = self.assertRaises(
            SystemExit,
            runner.run, ['--diff', 'path/to/previous.results', 'target'])

        self.assertEqual('path/to/previous.results', runner.diffOption)
        # Called once.
        self.assertEqual([True], showDiffResultsCalls)
        # Nothing in stderr or stdout.
        self.assertEqual('', self.outputStream.getvalue())
        self.assertEqual('', self.errorStream.getvalue())
        self.assertEqual(0, exitResult.code)


    def test_runDiffWarnings(self):
        """
        Exit with 1 when warnings are found in diff mode.
        """
        runner = self.makeRunner()
        runner.showDiffResults = lambda: 3

        exitResult = self.assertRaises(
            SystemExit,
            runner.run, ['--diff', 'path/to/previous.results', 'target'])

        self.assertEqual(1, exitResult.code)


    def test_showDiffResultsReadFail(self):
        """
        Show an error and exit with 1 when failing to read diff result file.
        """
        runner = self.makeRunner()
        runner.diffOption = 'no/such/file'

        result = runner.showDiffResults()

        self.assertEqual(1, result)
        self.assertEqual('', self.outputStream.getvalue())
        self.assertEqual(
            "Error: Failed to read result file 'no/such/file'.\n",
            self.errorStream.getvalue(),
            )

    def test_showDiffResultEmpty(self):
        """
        Return 0 when both sources are empty.
        """
        runner = self.makeRunner()
        runner.prepareDiff()
        runner._readDiffFile = lambda: ''

        result = runner.showDiffResults()

        self.assertEqual(0, result)
        self.assertEqual('', self.outputStream.getvalue())
        self.assertEqual('', self.errorStream.getvalue())


    def test_showDiffResultNoChanges(self):
        """
        Return 0 when both sources have same content.
        """
        runner = self.makeRunner()
        runner.prepareDiff()
        content = """
************* Module foo
W9001:  1,0: Missing copyright header
        """.strip()
        runner._readDiffFile = lambda: content
        runner.streamForDiff.write(content)

        result = runner.showDiffResults()

        self.assertEqual(0, result)
        self.assertEqual('', self.outputStream.getvalue())
        self.assertEqual('', self.errorStream.getvalue())


    def test_showDiffResultChanges(self):
        """
        Return 0 when both sources have same content.
        """
        runner = self.makeRunner()
        runner.prepareDiff()
        previous = """
************* Module foo
W9001:  1,0: Missing copyright header
        """.strip()
        new = """
************* Module foo
W9001:  1,0: Missing copyright header
W9001:  2,0: Missing copyright header
        """.strip()
        expectedOutput = """
************* Module foo
W9001:  2,0: Missing copyright header
""".lstrip()
        runner._readDiffFile = lambda: previous
        runner.streamForDiff.write(new)

        result = runner.showDiffResults()

        self.assertEqual(1, result)
        self.assertEqual(expectedOutput, self.outputStream.getvalue())
        self.assertEqual('', self.errorStream.getvalue())


    def test_formatWarnings(self):
        """
        Test for twistedchecker.core.runner.Runner.formatWarnings.
        """
        warnings = {
            "foo": {"W9001:  1,0: Missing copyright header", },
            "bar": {"W9002:  1,0: Missing a reference "
                    "to test module in header",
                    "C0111:  10,0: Missing docstring"
                   }
        }

        resultCorrect = """
************* Module bar
W9002:  1,0: Missing a reference to test module in header
C0111:  10,0: Missing docstring
************* Module foo
W9001:  1,0: Missing copyright header
        """.strip()

        result = Runner().formatWarnings(warnings)
        self.assertEqual(result, resultCorrect)


    def test_generateDiff(self):
        """
        Test for twistedchecker.core.runner.Runner.generateDiff.
        """
        oldWarnings = {
            "foo": {"W9001:  1,0: Missing copyright header"},
            "bar": {
                "W9002:  1,0: Missing a reference to test module in header",
                "C0111:  10,0: Missing docstring"
            }
        }

        newWarnings = {
            "foo": {
                "W9001:  1,0: Missing copyright header",
                "C0301: 10,0: Line too long"
            },
            "bar": {
                "W9002:  1,0: Missing a reference to test module in header",
                "C0111:  10,0: Missing docstring"
            },
            "baz": {
                "W9001:  1,0: Missing copyright header"
            }
        }

        diffCorrect = {
            "foo": {"C0301: 10,0: Line too long"},
            "baz": {"W9001:  1,0: Missing copyright header"}
        }

        # Make sure generated diff is correct.
        diff = Runner().generateDiff(oldWarnings, newWarnings)
        self.assertEqual(diff, diffCorrect)


    def test_getPathList(self):
        """
        Test for twistedchecker.core.runner.Runner.getPathList.
        """
        workingDir = os.getcwd()
        pathTwistedchecker = os.path.dirname(twistedchecker.__path__[0])
        inputList = [os.path.join("twistedchecker","functionaltests"),
                     "twistedchecker.core.util"]
        correctPaths = [os.path.join("twistedchecker","functionaltests"),
                        os.path.join("twistedchecker","core","util.py")]
        os.chdir(pathTwistedchecker)
        result = Runner().getPathList(inputList)
        # transform them to relative path.
        result = [os.path.relpath(path) for path in result]
        os.chdir(workingDir)

        self.assertEqual(result, correctPaths)


    def test_setNameExceptions(self):
        """
        Test for twistedchecker.core.runner.Runner.setNameExceptions.
        """
        pathTestFiles = createTestFilesForFindingExceptions(self.mktemp())
        self.clearOutputStream()
        runner = Runner()
        runner.setOutput(self.outputStream)
        # Set the reporter to C{twistedchecker.reporters.test.TestReporter}.
        runner.setReporter(TestReporter())
        # Limit messages.
        runner.linter.disable_noerror_messages()
        # Enable invalid function names.
        runner.linter.enable("C0103")
        # Enable invalid method names.
        runner.linter.enable("C9302")

        workingDir = os.getcwd()
        os.chdir(os.path.dirname(pathTestFiles))
        moduleName = os.path.basename(pathTestFiles)

        exitResult = self.assertRaises(SystemExit, runner.run, [moduleName])

        os.chdir(workingDir)
        predictResult = "7:C9302\n11:C0103\n14:C0103\n15:C9302\n"
        outputResult = self.outputStream.getvalue()
        self.assertEqual(outputResult, predictResult)
        self.assertEqual(16, exitResult.code)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Test cases for running the functional tests in
L{twistedchecker.functionaltests}.
"""

from functools import update_wrapper

import itertools
import os
import sys

from twisted.python.compat import _PY3, NativeStringIO
from twisted.python.reflect import filenameToModuleName
from twisted.trial import unittest

import twistedchecker
from twistedchecker.core.runner import Runner
from twistedchecker.reporters.test import TestReporter

try:
    from itertools import zip_longest
except:
    from itertools import izip_longest as zip_longest



def _partial2(wrapped, *partialArgs, **partialKwargs):
    """
    A custom implementation of L{functools.partial} which returns a function
    instead of a partial object.  This allows it to be assigned to a class
    attribute and used as an instance method.

    @see: U{https://twistedmatrix.com/trac/ticket/2645}
    @see: U{http://bugs.python.org/issue4331}

    @param wrapped: The function whose arguments will be partially applied.
    @type wrapped: L{callable}

    @param partialArgs: The positional arguments which will be applied to
        C{wrapped}.
    @type partialArgs: L{tuple}

    @param partialKwargs: The keyword arguments which will be applied to
        C{wrapped}.
    @type partialKwargs: L{dict}

    @return: A wrapper function which will apply the supplied arguments to
        C{wrapped} when it is called.
    @rtype: L{callable}
    """
    def wrapper(*args, **kwargs):
        args = args + partialArgs[len(args):]
        kwargs.update(partialKwargs)
        wrapped(*args, **kwargs)
    return update_wrapper(wrapper, wrapped)



def _formatResults(moduleName, expectedResult, actualResult):
    """
    Format the expected and actual results side by side for easy comparison.

    @param moduleName: The fully qualified module name.
    @type moduleName: L{str}

    @param expectedResult: The expected I{twistedchecker} output.
    @type expectedResult: L{str}

    @param actualResult: The actual I{twistedchecker} output.
    @type actualResult: L{str}

    @return: The expected and actual output formatted in two columns with
        headings.
    @rtype: L{str}
    """
    i = zip_longest(
        ['= Expected ='] + expectedResult,
        ['= Actual ='] + actualResult,
        fillvalue='')

    output = ['', moduleName]
    for col1, col2 in i:
        output.append(col1.ljust(20) + col2)
    return os.linesep.join(output)



def _parseLimitMessages(testFilePath):
    """
    Parse a test module file for a message control header.

    The first line of testfile should in format of:
    enable/disable: [Message ID], ...

    @param testFilePath: test module file to read.
    @type testFilePath: L{str}

    @return: A 2-tuple of (C{action}, C{messages}) where C{action} is either
        C{"enable"} or C{"disable"} and C{messages} is a list of
        I{twistedchecker} message IDs.
    @rtype: L{tuple}
    """
    with open(testFilePath) as f:
        firstline = f.readline()
    if "enable" not in firstline and "disable" not in firstline:
        # Could not find enable or disable messages
        return
    action, messages = firstline.lstrip("#").strip().split(":", 1)
    messages = [msgid for msgid in messages.strip().split(",") if msgid]
    action = action.strip()

    return action, messages



def _setLinterLimits(linter, action, messages):
    """
    Enable or disable the reporting of certain linter messages.

    @param linter: The linter whose
        L{twistedchecker.checkers.pep8format.PEP8Checker} will be enabled.
    @type linter: L{pylint.lint.PyLinter}

    @param action: Either C{"enable"} or C{"disable"}.
    @type action: L{str}

    @param messages: A list of twistedchecker message IDs to be enabled or
        disabled.
    @type messages: L{list} of L{str}
    """
    if action == "enable":
        # Disable all other messages
        linter.disable_noerror_messages()

    messageModifier = getattr(linter, action)
    for msgid in messages:
        messageModifier(msgid)



def _runTest(testCase, testFilePath):
    """
    Run a functional test.

    @param testCase: The test case on which to call assertions.
    @type testCase: L{unittest.TestCase}

    @param testFilePath: The path to the module to test.
    @type testFilePath: L{str}
    """
    pathResultFile = testFilePath.replace(".py", ".result")
    moduleName = filenameToModuleName(testFilePath)
    outputStream = NativeStringIO()

    runner = Runner()
    runner.allowOptions = False
    runner.setOutput(outputStream)
    runner.setReporter(TestReporter())

    limits = _parseLimitMessages(testFilePath)
    if limits is not None:
        action, messages = limits
        _setLinterLimits(runner.linter, action, messages)

    exitCode = None
    try:
        runner.run([moduleName])
    except SystemExit as error:
        exitCode = error.code

    # Check the results
    with open(pathResultFile) as f:
        expectedResult = sorted(f.read().strip().splitlines())

    outputResult = sorted(outputStream.getvalue().strip().splitlines())

    try:
        testCase.assertEqual(expectedResult, outputResult)
    except unittest.FailTest:
        testCase.fail(_formatResults(moduleName, expectedResult, outputResult))

    if not expectedResult:
        testCase.assertEqual(0, exitCode)
    else:
        testCase.assertNotEqual(0, exitCode)



def _testNameFromModuleName(moduleName):
    """
    Mangle a module name so it can be used as a test function name.

    @param moduleName: The qualified module name.
    @type moduleName: L{str}

    @return: The test name derived from the supplied C{moduleName}.
    @rtype: L{str}
    """
    return 'test_' + moduleName.replace('.', '_')



def _testModules():
    """
    Discover all the functional test modules.

    Modules whose name begin with an underscore are ignored.

    @return: An iterator of test module file paths.
    @rtype: L{iter}
    """
    pathTestModules = os.path.join(twistedchecker.abspath, "functionaltests")
    for root, dirs, files in os.walk(pathTestModules):
        for testfile in files:
            if testfile.startswith("_"):
                continue
            if testfile.endswith(".py"):
                if _PY3 and testfile == "test_class_name.py":
                    # This one doesn't work in tox in py3 for some reason.
                    continue
                else:
                    yield os.path.join(twistedchecker.abspath, root, testfile)



def _testsForModules(testModules):
    """
    Return a dictionary of test names and test functions.

    @param testModules: An iterable list of functional test module names.
    @type testModules: L{iter}

    @return: A dictionary of test functions keyed by test name.
    @rtype: L{dict} of (L{str}, L{callable})
    """
    t = []
    for modulePath in testModules:
        moduleName = filenameToModuleName(modulePath)
        t.append(
            (_testNameFromModuleName(moduleName),
             _partial2(_runTest, testFilePath=modulePath))
        )
    return dict(t)



# Create a TestCase class using a dictionary of dynamically generated test
# methods.
FunctionalTests = type(
    "FunctionalTests",
    (unittest.TestCase,),
    _testsForModules(_testModules())
)



import os

from twisted.python.compat import NativeStringIO
from twisted.trial import unittest

import twistedchecker
from twistedchecker.core.runner import Runner



class LimitedReporterTestCase(unittest.TestCase):
    """
    Test for twistedchecker.reporter.limited.LimitedReporter.
    """

    def test_reporter(self):
        """
        Test for LimitedReporter.
        Use test file of indentation to test
        whether limited messages are returned when using LimitedReporter.

        Complete run on the test file will return two warnings:
        W0311 and W0312, but with limited report it returns only one.
        """
        moduleTestIndentation = "twistedchecker.functionaltests.indentation"
        pathTestIndentation = os.path.join(twistedchecker.abspath,
                                      "functionaltests", "indentation.py")
        # assert the test file exists
        self.assertTrue(os.path.exists(pathTestIndentation))
        streamTestResult = NativeStringIO()
        runner = Runner()
        runner.setOutput(streamTestResult)
        # defaultly, runner will use LimitedReporter as its output reporter
        # set allowed messages for it
        runner.linter.reporter.messagesAllowed = set(["W0311"])

        exitResult = self.assertRaises(
            SystemExit, runner.run, [moduleTestIndentation])

        # check the results to see only W0311 is reported
        resultTest = streamTestResult.getvalue()
        self.assertTrue("W0311" in resultTest)
        self.assertTrue("W0312" not in resultTest)
        self.assertEqual(4, exitResult.code)
